---
title: Security
permalink: /security
category: Cube.js Backend
menuOrder: 4
---

Cube.js uses [JSON Web Tokens (JWT)](https://jwt.io/) for requests' authorization and also for passing
additional user context, which could be used in the [USER_CONTEXT](cube#context-variables-user-context) object in the Data
Schema.

Cube.js tokens are designed to work in micro services environment.
Typical use case would be:

1. There's web server that serves HTML with JS client code that calls cube.js.
2. Web server should generate expirable cube.js tokens and incorporate them as part of HTML or send it over XHR request in exchange of session cookie or other security credentials.
3. JS Client code uses token to call cube.js server API.

If you are using [REST API](rest-api) you need pass API Token via the Authorization Header.
Cube.js Javascript client accepts auth token as a first argument to [cubejs(authToken, options) function](@cubejs-client-core#cubejs).

**In the development environment the token is not required for authorization**, but
you can still use it to [pass a security context](security#security-context).

## Generating Token

Auth token is generated based on your API secret. Cube.js CLI generates API Secret on app creation and saves it in `.env` file as `CUBEJS_API_SECRET` variable.

You can generate two types of tokens:
- Without security context. It implies same data access permissions for all users.
- With security context. User or role-based security models can be implemented using this approach.

_It is considered best practice to use `exp` expiration claim to limit life time of your public tokens.
[Learn more at JWT docs](https://github.com/auth0/node-jsonwebtoken#token-expiration-exp-claim)._

### Using CLI

You can use a Cube.js CLI [token](reference#token) command to generate an API token.

```bash
$ cubejs token -e TOKEN-EXPIRY -s SECRET -p FOO=BAR
```

However it is handy to create an API token with CLI command for testing
purposes, we strongly recommend to programmatically generate tokens in production.

### Programmatically

You can find a library for JWT generation for your programming language [here](https://jwt.io/#libraries-io).

Below you can find an example on how to generate an API token in Node.js with
`jsonwebtoken` package:

```javascript
const jwt = require('jsonwebtoken');
const CUBE_API_SECRET='secret';

const cubejsToken = jwt.sign({}, CUBE_API_SECRET, { expiresIn: '30d' });
```

## Security Context

Security context can be provided by passing `u` param for payload.
For example if you want to pass user id in security context you can create token with payload:
```json
{
  "u": { "id": 42 }
}
```

In this case `{ "id": 42 }` object will be accessible as [USER_CONTEXT](cube#context-variables-user-context) in the Cube.js Data Schema.

Consider the following example. We want to show orders only for
customers, who owns these orders. `orders` table has a `user_id` column, which we
can use to filter the results.

```javascript
cube(`Orders`, {
  sql: `SELECT * FROM public.orders WHERE ${USER_CONTEXT.id.filter('user_id')}`,

  measures: {
    count: {
      type: `count`
    }
  }
});
```

Now, we can generate an API Token with user ID:

```javascript
const jwt = require('jsonwebtoken');
const CUBE_API_SECRET='secret';

const cubejsToken = jwt.sign({ u: { id: 42 } }, CUBEJS_API_SECRET, { expiresIn: '30d' });
```

Using this token we can sign our request to Cube.js Backend.

```bash
curl \
 -H "Authorization: eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1Ijp7ImlkIjo0Mn0sImlhdCI6MTU1NjAyNTM1MiwiZXhwIjoxNTU4NjE3MzUyfQ._8QBL6nip6SkIrFzZzGq2nSF8URhl5BSSSGZYp7IJZ4" \
 -G \
 --data-urlencode 'query={"measures":["Orders.count"]}' \
 http://localhost:4000/cubejs-api/v1/load
````

And the following SQL will be generated by Cube.js.

```sql
SELECT
  count(*) "orders.count"
  FROM (
    SELECT * FROM public.orders WHERE user_id = 42
  ) AS orders
LIMIT 10000
```
